#! /usr/bin/perl
###############################################################################
#
# Ferret user interface program
#
# Written by Brian White <bcwhite@verisim.com>
# Copyright (c) 1996 by Verisim, Inc.
#
###############################################################################



=head1 NAME

ferret - "Ferret" search engine command-line interface

=cut



# Load the big guns...
use lib "blib/lib";
use Ferret;



###############################################################################
#
# Program-wide variable declarations (and defaults)
#
$debug=0;
$search;
$action= '';
$imode = 1;
@parms = ();
%opts  = (
	addprefix	=> '',
	doctype		=> '',
	filter      => '',
	index		=> 'ferret.index',
	lines       => '',
	stripprefix	=> '',
	summary		=> '250',
);
%flags = (
	filter		=> 0,
	force		=> 0,
	help		=> 0,
	notitles	=> 0,
	summaries   => 0,
);

$opts{index} = $ENV{FERRET_INDEX} if $ENV{FERRET_INDEX};

if ($debug) {
	use FileHandle;
	autoflush STDOUT 1;
}



###############################################################################
#
# Usage message
#
my $usage = qq"
Use: $0 [--index=<index-file>] <action> [action parameters] [...]

Actions:
     addfile        adds named file(s) to index
     addstoppers    adds to list of stopper (non-indexed) words
     commonwords    reports words that exist in many documents
     query          searches for specified words (put in 'quotes')
     removefile     removes named file(s) from index
     removestoppers removes from list of stopper (non-indexed) words
     setoption      set specified options
     shrink         reduce index size after adds or removes
     unsetoption    unset specified options

Type \"$0 --help <action>\" for more information.

";


=head1 SYNOPSIS

B<ferret> [B<--index=><I<file>>] <I<action>> [I<action parms>] [...]

Actions:
    B<addfile>        adds named file(s) to index
    B<addstoppers>    adds stopper (non-indexed) words
    B<commonwords>    report words existing in many files
    B<query>          searches for specified words
    B<removefile>     removes named file(s) from index
    B<removestoppers> removes stopper (non-indexed) words
    B<setoption>      set specified options
    B<shrink>         reduce index size after adds/removes
    B<unsetoption>    unset specified options

=head1 DESCRIPTION

The B<ferret> command provides a convienient interface to update or perform
maintenence on an index.

=head2 General Options

The following option is applicable to I<all> actions.

=over 4

=item B<--index=><I<file>>

This option specifies which index file is to be used.  If this option is
omitted, it will default to the value of the environment variable
C<FERRET_INDEX> or "F<./ferret.index>" if C<FERRET_INDEX> is not set.

=back

=cut


###############################################################################
#
# Work ("action") subroutines follow
#
###############################################################################



###############################################################################
#
# Utility Subroutines
#

sub BadFileTypes {
	my($fref) = @_;
	my $temp;

	$temp = pack("N",hex("13579ACE"));
	return "database file" if $$fref =~ m/^\Q$temp\E/;

	$temp = pack("V",hex("13579ACE"));
	return "database file" if $$fref =~ m/^\Q$temp\E/;

	return;
}



###############################################################################
#
# Action:  ADDFILE <files>
#

=head1 AddFile

Use: B<ferret> [...] B<addfile> [B<--addprefix=><I<path>>] [B<--stripprefix=><I<regex>>]
[B<--notitles>] [B<--summary=><I<size>>] [B<--lines=><I<number>>] [B<--doctype=><I<type>>]
[B<--filter=>'<I<program & args>>'] [B<--force>] <I<file-to-add>> [...]

Options:
    B<addprefix>    add to each filename before indexing
    B<doctype>      process as documents of this type
    B<filter>       run raw data through this program
    B<force>        add regardless of modification time
    B<lines>        limit summary to this number of lines
    B<notitles>     supress titles for indexed documents
    B<stripprefix>  strip from each filename before indexing
    B<summary>      limit summary to this number of bytes

Creating an index of a local repository of files is easily done with the
"ferret addfile" maintenence command.  A command such as

    find /usr/local/repository -type f -print | xargs -n 500 \
         ferret --index=/usr/local/index addfile

will create an index of the text of all files under the 'repository' directory
in /usr/local.  The 'ferret' program is smart enough to recognize standard
extensions and handle them accordingly.  Compressed files are uncompressed
before being indexed.  Recognized file types are HTML, Code (C/C++, Perl,
shell-script), and MIF (Maker Interchange Format).  Any unknown types are
assumed to be plain text.  This is a safe fall-back since any binary codes are
stripped out before any document is added to the index.

After calling "ferret addfile" multiple times, as 'xargs' is bound to do, it
might help to run the command

    ferret --index=/usr/local/index shrink

to remove any obsolete data and pack the index into as small a file as
possible.

=head2 AddFile-specific Parameters

The following parameters are specific to the "addfile" action.

=over 4

=item B<--addprefix=><I<string>>

This option will cause B<AddFile> to prepend a given string to each filename
after it has been read from disk but before it is stored in the index.  This
could be used to change the pathnames so they will work when seen by the query
program.  Note that this happens after any prefix has been stripped (see
--stripprefix).

If not specified, nothing will be prepended to the pathnames.

Example:  C<--addprefix="/mnt/">

=item B<--doctype=><I<type>>

This option will cause B<AddFile> to add all listed document as being of the
specified type.  This is best used when adding difficult to recognize types
such as code for various languages.  Valid types are HTML, Code, MIF, and
Text.  Unlike many options, these type name I<are> case sensitive.

If not specified, B<AddFile> will try to dynamically determine the type for
each file on an individual basis based on the filename extension and its
content.

Example:  C<--doctype=Code>

=item B<--filter=>'<I<program & args>'>

This option will cause B<AddFile> to run each file through the given program
(with the specified arguments) before indexing the data.  This program must
accept the file on STDIN and write its output to STDOUT.

If not specified, B<AddFile> will dynamically determine what filters should be
run based on the file's extension.  Currently supported extensions are B<.gz>
(gzip) and B<.Z> (compress).

Example:  C<--filter='gzip -d | pod2text'>

=item B<--force>

This option will cause B<AddFile> to add every file on the command line even
if the last modification date of those files is older than when they were last
added to the index.

Example:  C<--filter='gzip -d | pod2text'>

=item B<--lines=><I<number>>

This option will cause B<AddFile> to limit summaries to a maximum of this
number of lines.  This option will not affect the summaries generated for HTML
documents because there is no way to count lines until the document is
actually displayed.

If not specified, a maximum of 5 lines will be generated.

Example:  C<--lines=10>

=item B<--notitles=>

This option will cause B<AddFile> to supress storing titles for the documents
it indexes.  Display/Query programs can use titles to better show the user
what matches have been found.

If not specified, titles will be stored.

Example:  C<--notitles>

=item B<--stripprefix=><I<regex>>

This option will cause B<AddFile> to remove the specified pattern from each
filename after it has been read from disk but before it is stored in the
index.  This could be used to change the pathnames so they will work when seen
by the query program.  Note that this happens after before prefix is added
(see --addprefix).  The regex is evaluated (case insensitive) by Perl and thus
can use all of the Perl extensions.

If not specified, nothing will be stripped from the pathnames.

Example:  C<--stripprefix=".*/">

=item B<--summary=><I<size>>

This option will cause B<AddFile> to limit the size summaries can be.  This is
most useful for HTML documents since straight text documents are better
restricted by number of lines (--lines option).

If not specified, the maximum summary size is 250 bytes.

Example:  C<--summary=500>

=back

=cut

sub AddFile {
	if ($flags{help} || @parms < 1) {
		print STDERR "\nUse: $0 [...] addfile [--addprefix=<path>] [--stripprefix=<path>]\n";
		print STDERR "     [--notitles] [--summary=<size>] [--lines=<number>] [--doctype=<type>]\n";
		print STDERR "     [--filter='<program & args>'] <file-to-add> [...]\n";
		print STDERR "\nOptions:\n";
		print STDERR "     addprefix    add to each filename after loading but before indexing\n";
		print STDERR "     doctype      process as documents of this type:  HTML, Text, Code\n";
		print STDERR "     filter       run row data through this program (+args) before indexing\n";
		print STDERR "     force        add all files regardless of last modification date\n";
		print STDERR "     lines        limit summary to this number of lines in length\n";
		print STDERR "     notitles     supress storing titles for indexed documents\n";
		print STDERR "     stripprefix  strip from each filename after loading but before indexing\n";
		print STDERR "     summary      store maximum of this number of bytes as a document summary\n";
		die "\n";
	}

	my($filename,$ignored);

	foreach $filename (@parms) {
		my($filter,$fixed,$filtname,$data,$title,$summary,$dtype);
		$summary= "";

		$fixed = $filename;
		$fixed =~s!^$opts{stripprefix}!! if $opts{stripprefix};
		$fixed = $opts{addprefix} . $fixed;

		my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
		   $atime,$mtime,$ctime,$blksize,$blocks) = stat($filename);
		if ($search->DocumentTimestamp($fixed) > $mtime) {
			if ($flags{force}) {
				print "File '$filename' is unchanged since last added -- forced\n";
			} else {
				print "File '$filename' is unchanged since last added -- ignored\n";
				$ignored = 1;
				next;
			}
		}

		if ($opts{filter}) {
			$filter = $opts{filter};
			$fixed  = $filename;
		} else {
			$filter = Ferret::StdExtFilters($filename,\$filtname);
		}
		$data   = Ferret::LoadFile($filename,$filter);
		$dtype  = $opts{doctype};

		my $badfile = BadFileTypes(\$data);
		if ($badfile) {
			if ($flags{force}) {
				print "Warning: forced indexing of $badfile '$filename'\n";
			} else {
				print "Warning: not indexing $badfile '$filename'\n";
				next;
			}
		}

		unless ($dtype) {
			$dtype = "HTML"	if ($filtname =~ m/\.html?$/i || $data =~ m/^\s*<html>/si);
			$dtype = "Code" if ($filtname =~ m/\.(h|hh|hpp|h\+\+|c|cc|cpp|c\+\+)$/i || $data =~ m/^\#\!/);
			$dtype = "MIF"	if ($filtname =~ m/\.mif$/i || $data =~ m/^<MIFFile[\s\d\.]*>/);
			$dtype = "Mail"	if (!$dtype && $data =~ m/^from /i);
			$dtype = "Text" unless $dtype;
		}

		print "Adding '$filename' ";
		print "as '$fixed' " if $filename ne $fixed;
		print "($dtype) ...\n";
		eval 'Ferret::Strip' . $dtype . '(\$data,\$title,\$summary,$opts{summary})';
		die "$0: Unknown document type '$dtype'\n" if $@;
		Ferret::LimitLineCount(\$summary,$opts{lines}) if $opts{lines} && $dtype ne "HTML";
		Ferret::MakeHTMLSummary(\$summary, $dtype);
		$search->AddDocument($fixed,$data);
		$search->DBPutTitle("$fixed",$title)		if $title   && !$flags{notitle};
		$search->DBPutSummary("$fixed",$summary)	if $summary &&  $opts{summary};
	}

	if ($ignored) {
		print '(use the "--force" option to force adding of unmodified files)',"\n";
	}
}



###############################################################################
#
# Action:  ADDSTOPPERS <words>
#

=head1 AddStoppers

Use: B<ferret> [...] B<addstoppers> <I<word>> [...]

B<Ferret> uses a concept of "stoppers" to reduce the amount of information it
stores in its index.  These non-content words (eg. "the", "where",
"shouldn't", etc.) can safely be removed from most indexes because they don't
actually mean anything.  A default list is provided by B<Ferret>, but can
easily be added to on a per-index basis using this function.  Once added,
these words are remembered until the index is deleted or they are removed
using the B<removestoppers> command.

Once added, that word and all the data associated with it is forever removed
from the index, thus reducing the size of the index file.  Stoppers are also
removed from queries in an intelligent manner so most users will never notice
that some words were completely ignored.

=cut

sub AddStoppers {
	if ($flags{help} || @parms < 1) {
		print STDERR "\nUse: $0 [...] addstoppers <word> [...]\n";
		die "\n";
	}

	$search->AddStoppers(@parms);
}



###############################################################################
#
# Action:  COMMONWORDS <min> [max]
#

=head1 CommonWords

Use: B<ferret> [...] B<commonwords> <I<min>> [I<max>]

Both I<min> and I<max> can be either a number between 0 and 1 to indicate a a
frequency (eg. 0.90 = 90% of all documents) or a whole number greater than 1
to indicate an exact number of documents.  Running "commonwords 0.90" will
return a list of words in more than 90% of the documents.  This is useful for
determining words to be added to the "stopper" list.  (see: "addstoppers" and
"removestoppers") Running "commonwords 0.00 0.10" will list all the words that
are in less than 10% of all documents.

In general, using B<addstoppers> to avoid indexing words that appear in more than
50% of all documents will not degrade query performance and can reduce the index
size by a siginificant amount.

=cut

sub CommonWords {
	if ($flags{help} || @parms < 1 || @parms > 2) {
		print STDERR "\nUse: $0 [...] commonwords <min> [max]\n";
		print STDERR "\nBoth <min> and <max> can be either a number between 0 and 1 to indicate a\n";
		print STDERR "a frequency (eg. 0.90 = 90% of all documents) or a whole number greater than\n";
		print STDERR "1 to indicate an exact number of documents.  Running \"commonwords 0.90\"\n";
		print STDERR "will return a list of words in more than 90% of the documents.  This is useful\n";
		print STDERR "for determining words to be added to the \"stopper\" list.  (see: \"addstoppers\"\n";
		print STDERR "and \"removestoppers\")  Running \"commonwords 0.00 0.10\" will list all the words\n";
		print STDERR "that are in less than 10% of all documents.\n";
		die "\n";
	}

	my $words;

	if (@parms == 1) {
		$words = $search->CommonWords($parms[0]);
	} else {
		$words = $search->CommonWords($parms[0],$parms[1]);
	}

	my @words = sort(split(/\n/,$words));

	print "common words: @words\n";
}



###############################################################################
#
# Action:  QUERY 'query-string'
#

=head1 Query

Use: B<ferret> [...] B<query> [B<--summaries>] '<I<query-string>>'

Options:
    B<summaries>    display summaries for all matches found

The query string should be enclosed in single quotes so that double quotes can
be passed as part of the query.  Any single quotes in the query (for
apostrophes in contractions) will have to be escaped with a backslash.

For more information on the format of the query, see the I<ferret(3)> man page.

=cut

sub Query {
	if ($flags{help} || @parms != 1) {
		print STDERR "\nUse: $0 [...] query [--summaries] 'query-string'\n";
		print STDERR "\nOptions:\n";
		print STDERR "     summaries    display summaries for all matches found\n";
		print STDERR "\nThe query string should be enclosed in single quotes so that double quotes\n";
		print STDERR "can be passed as part of the query.  Any single quotes in the query (for\n";
		print STDERR "apostrophes in contractions) will have to be escaped with a backslash.\n";
		die "\n";
	}

	my @results = $search->Query($parms[0]);
	die "$@\n" if $@;

	print "Score Match\n~~~~~ ~~~~~\n";
	foreach (@results) {
		my($score,$match) = (m/^(\d+) (.*)$/);
		printf " %3d  %s\n",$score,$match;

		if ($flags{summaries}) {
			my $summary = $search->DBGetSummary($match);

			if ($summary) {
				$summary =~ s/(^|\n)/$1      /g;
				print $summary,"\n\n";
			}
		}
	}
}



###############################################################################
#
# Action:  REMOVEFILE <files>
#

=head1 RemoveFile

Use: B<ferret> [...] B<removefile> [B<--addprefix=><I<path>>] [B<--stripprefix=><I<regex>>]
<I<file-to-remove>> [...]

Options:
    B<addprefix>    add to each filename before indexing
    B<stripprefix>  strip from each filename before indexing

The B<RemoveFile> action will remove a document from the index.  Note that the
name provided as I<file-to-remove> must match the indexed name -- I<not> the
file name.  The indexed name is the file name with any prefixes stripped and
added.

The following parameters are specific to the "removefile" action and behave
identically to those of the "addfile" action.  See L</AddFile> S< > for more
information.

=cut

sub RemoveFile {
	if ($flags{help} || @parms < 1) {
		print STDERR "\nUse: $0 [...] removefile [--addprefix=<path>] [--stripprefix=<path>]\n";
		print STDERR "     <file-to-remove> [...]\n";
		print STDERR "\nOptions:\n";
		print STDERR "     addprefix    add to each filename after loading but before indexing\n";
		print STDERR "     stripprefix  strip from each filename after loading but before indexing\n";
		die "\n";
	}

	my($filename);

	foreach $filename (@parms) {
		my($filter,$filtfile,$fixed,$removed);
		$filter = Ferret::StdExtFilters($filename,\$filtfile);
		$fixed  =~s!^$opts{stripprefix}!! if $opts{stripprefix};
		$fixed  = $opts{addprefix} . $fixed;

		print "Removing '$filename' ";
		print "as '$fixed' " if $filename ne $fixed;
		print "...\n";
		$removed = $search->RemoveDocument($fixed);
		print "Warning: document '$fixed' was not in index\n" unless $removed;
		$search->DBDelSummary("$fixed");
	}
}



###############################################################################
#
# Action:  REMOVESTOPPERS <words>
#

=head1 RemoveStoppers

Use: B<ferret> [...] B<removestoppers> <I<word>> [...]

If there are "stopper" words that should be indexed, simply remove those words
from the list using this command.  Note that 1 and 2 letter words are
I<always> considered stoppers unless turned off using the B<nostoppers> option
(see L</NoStoppers>).

If a stopper is removed from the list, all future documents will be indexed on
that word.  Documents already indexed, however, will not be until next updated.

=cut

sub RemoveStoppers {
	if ($flags{help} || @parms < 1) {
		print STDERR "\nUse: $0 [...] removestoppers <words> [...]\n";
		die "\n";
	}

	$search->RemoveStoppers(@parms);
}



###############################################################################
#
# Action:  SETOPTION <options>
#

=head1 SetOption / UnsetOption

Use: B<ferret> [...] B<setoption>|B<unsetoption> <I<option>> [...]

Options:
    B<tiny>         make index as small as possible
    B<nostoppers>   don't remove any stopper words

This action will set index-specific options.  These options will remain in
effect until the index file is deleted or until unset with the "UnsetOption"
action.

=over 4

=item B<Tiny>

A "tiny" database can use significantly less disk space but does so at the
cost of all proximity searches, including phrases.

If this option is set when documents already exist in the index, there will be
a delay while all of the extraneous information is removed.  Be sure to run
the the B<shrink> action afterwards to reclaim disk space back from the index
file.

B<Important:>  This option I<cannot> be unset!  The index file must be erased and
regenerated from scratch.

=item B<NoStoppers>

If it is important to be able to search for I<all> words, the "nostoppers"
option will ensure that every word is indexed.  While this can dramatically
increase the size of an index file, it does ensure that, for example,
searching Shakespeare for "to be or not to be" (a phrase composed entirely of
stoppers) will succeed.  (Hamlet, Act III, Scene I)

Note that documents are indexed and queries are run based on the current state
of this option, which could cause some confusion to the user if changed
without reindexing all of the stored documents.  If this option is to be used,
it should probably be set immediately after opening the index for the first
time and never unset.

=back

=cut

sub SetOption {
	if ($flags{help} || @parms < 1) {
		print STDERR "\nUse: $0 [...] setoption <option> [...]\n";
		print STDERR "\nOptions:\n";
		print STDERR "     tiny         make index as small as possible (loses proximity searching)\n";
		print STDERR "     nostoppers   don't remove any stopper (non-content) words\n";
		die "\n";
	}

	foreach (@parms) {
		my $option=0;

		/^tiny$/i		and $option=&Ferret::OPT_TINY;
		/^nostoppers$/i	and $option=&Ferret::OPT_NOSTOPPERS;

		print "Setting option '$_' ...\n";
		$search->SetOption($option);
	}
}



###############################################################################
#
# Action:  SHRINK
#

=head1 Shrink

Use: B<ferret> [...] B<shrink>

This action will reclaim all unused space from the index.  While that space
would be reused during future B<addfile> actions, it will only be returned to
the file system with this call.

=cut

sub Shrink {
	if ($flags{help} || @parms != 0) {
		print STDERR "\nUse: $0 [...] shrink\n";
		die "\n";
	}

	print "Shrinking index...\n";
	$search->Shrink();
}



###############################################################################
#
# Action:  UNSETOPTION <options>
#
sub UnsetOption {
	if ($flags{help} || @parms < 1) {
		print STDERR "\nUse: $0 [...] unsetoption <option> [...]\n";
		print STDERR "\nOptions:\n";
		print STDERR "     tiny         make index as small as possible (cannot be unset)\n";
		print STDERR "     nostoppers   don't remove any stopper (non-content) words\n";
		die "\n";
	}

	foreach (@parms) {
		my $option=0;

		/^tiny$/i		and $option=&Ferret::OPT_TINY;
		/^nostoppers$/i	and $option=&Ferret::OPT_NOSTOPPERS;

		print "Unsetting option '$_' ...\n";
		$search->UnsetOption($option);
	}
}



###############################################################################
#
# Main program follows
#
###############################################################################


###############################################################################
#
# Parse command line arguments
#
die $usage unless @ARGV > 0;

foreach (@ARGV) {
	if (/^--(.*?)=(.*)$/) {
		if (defined $opts{$1}) {
			$opts{$1} = $2;
		} else {
			die "$0: Unknown option '--$1'\n" . $usage;
		}
	} elsif (/^--(.*?)$/) {
		if (defined $flags{$1}) {
			$flags{$1} = 1;
		} else {
			die "$0: Unknown flag '--$1'\n" . $usage;
		}
	} elsif ($action) {
		push @parms,$_;
	} else {
		$action = $_;
		if (/^addfile$/)		{ 			next; }
		if (/^addstoppers$/)	{ 			next; }
		if (/^commonwords$/)	{ $imode=0;	next; }
		if (/^query$/)			{ $imode=0;	next; }
		if (/^removefile$/)		{ 			next; }
		if (/^removestoppers$/)	{ 			next; }
		if (/^setoption$/)		{ 			next; }
		if (/^shrink$/)			{ 			next; }
		if (/^unsetoption$/)	{ 			next; }

		die "$0: Unknown action '$_'\n" . $usage;
	}
}

die $usage unless $action;



###############################################################################
#
# Create a Ferret and open the specified index (only if not in "help" mode)
#
unless ($flags{help}) {
	$search = new Ferret;
	if ($imode) {
		$search->Update($opts{index});
	} else {
		$search->Open($opts{index});
	}
}



###############################################################################
#
# Call the appropriate routine for the requested action
#
AddFile()		if ($action eq "addfile");
AddStoppers()	if ($action eq "addstoppers");
CommonWords()	if ($action eq "commonwords");
Query()			if ($action eq "query");
RemoveFile()	if ($action eq "removefile");
RemoveStoppers()if ($action eq "removestoppers");
SetOption()		if ($action eq "setoption");
Shrink()		if ($action eq "shrink");
UnsetOption()	if ($action eq "unsetoption");



###############################################################################
#
# Put things away when done...
#
print "\nWriting index...\n" if $imode;
$search->Close();



=head1 COPYRIGHT

B<Ferret> is copyright (c) 1996 by Verisim, Inc.

=head1 SEE ALSO

B<ferret>(3)

For more information, join the B<Ferret> mailing list by mailing the command
"subscribe ferret-list" to the address: C<ferret-list-requst@verisim.com>

=head1 BUGS

Report bugs to: C<ferret-bugs@verisim.com> S< > Please include the version of
B<Ferret> being run, a detailed description of the problem, and how to
reproduce it.  Thank you!

=cut
